import math
import pytest

import polars as pl
from tsfx import (
    ExtractionSettings,
    FeatureSetting,
    extract_features,
    DynamicGroupBySettings,
)


def test_empty_df():
    df = pl.DataFrame({"id": [], "val": []})
    df = df.select([pl.col("id").cast(pl.Utf8), pl.col("val").cast(pl.Float64)])
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df.lazy(), opts)
    assert fdf.is_empty()


def test_unit_length_df():
    df = pl.DataFrame({"id": ["a"], "val": [1.0]}).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    assert fdf.shape[0] == 1


def test_only_nan_group_dropped():
    df = pl.DataFrame({"id": ["a", "b", "c", "c"], "val": [1.0, 2.0, None, None]})
    df = df.select([pl.col("id").cast(pl.Utf8), pl.col("val").cast(pl.Float64)])
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df.lazy(), opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("id").to_list() == ["a", "b", "c"]

def test_only_nan_group_dropped_with_date():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "c", "c", "c"],
            "time": [
                "2001-01-01",
                "2002-01-01",
                "2003-01-01",
                "2001-01-01",
                "2002-01-01",
                "2003-01-01",
                "2001-01-01",
                "2002-01-01",
                "2003-01-01",
            ],
            "val": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0, None, None, None],
            "value": [4.0, 5.0, 6.0, 6.0, 5.0, 4.0, 4.0, 5.0, 6.0],
        },
    ).lazy()

    dyn_settings = DynamicGroupBySettings(
        time_col="time",
        every="1y",
        period="1y",
        offset="0y",
        datetime_format="%Y-%m-%d",
    )
    settings = ExtractionSettings(
        grouping_col="id",
        value_cols=["val", "value"],
        feature_setting=FeatureSetting.Efficient,
        dynamic_settings=dyn_settings,
    )
    fdf = extract_features(df, settings)
    assert fdf.get_column("id").to_list() == ["a", "a", "a", "b", "b", "b", "c", "c", "c"]
    assert fdf["val__mean"].fill_nan(None).to_list()[0] == 1.0
    assert fdf["val__mean"].fill_nan(None).to_list()[1] == 2.0
    assert fdf["val__mean"].fill_nan(None).to_list()[2] == 3.0
    assert fdf["val__mean"].fill_nan(None).to_list()[-1] == None

def test_long_constant_df():
    N = 100_000
    df = pl.DataFrame({"id": ["a"] * (N + 1) + ["b"] * (N + 1), "val": [0.0] * N + [1.0] + [1.0] * N + [2.0]}).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("id").to_list() == ["a", "b"]
    assert fdf.get_column("val__mean").to_list() == pytest.approx([0.0, 1.0], abs=1e-4)



def test_nan_df():
    df = pl.DataFrame({"id": ["a", "b", "c", "c"], "val": [1.0, 2.0, 1.0, None]}).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("id").to_list() == ["a", "b", "c"]
    assert fdf.get_column("val__mean").to_list() == [1.0, 2.0, 1.0]


def test_nan_df2():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "c", "c"] + ["d"] * 10,
            "val": [1.0, 2.0, 1.0, None, None, 1.0] + [None] * 8 + [1.0, 1.0],
            "val2": [1.0, None, None, 1.0, None, 1.0] + [1.0] * 8 + [None, None],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("id").to_list() == ["a", "b", "c", "d"]
    assert fdf.get_column("length").to_list() == [2.0, 1.0, 1.0, 2.0]


def test_length():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 1.0, 2.0, 3.0, 1.0, 1.0, 1.0, 1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("length").to_list() == [1.0, 2.0, 3.0, 4.0]


def test_sum_values():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 1.0, 2.0, 3.0, 1.0, 1.0, 1.0, 1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__sum_values").to_list() == [1.0, 3.0, 6.0, 4.0]


def test_minimum():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 1.0, 2.0, 3.0, 1.0, 1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__minimum").to_list() == [1.0, 1.0, 1.0, -1.0]


def test_maximum():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 1.0, 2.0, 3.0, 1.0, 1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__maximum").to_list() == [1.0, 2.0, 3.0, 1.0]


def test_absolute_energy():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, -2.0, 1.0, -2.0, 3.0, 1.0, 1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__absolute_energy").to_list() == [1.0, 5.0, 14.0, 4.0]


def test_median1():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 1.0, 2.0, 3.0, 1.0, 1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__median").to_list() == [1.0, 1.5, 2.0, 0.0]


def test_median2():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 1.0, 2.0, 3.0, 1.0, 1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__median").to_list() == [1.0, 1.5, 2.0, 0.0]


def test_linear_trend_intercept():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 1.0, 2.0, 3.0, 2.0, 1.0, -1.0, -1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__linear_trend_intercept").to_list() == pytest.approx(
        [1.0, 1.0, 3.0, -1.0], abs=1e-6,
    )


def test_linear_trend_slope():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 1.0, 2.0, 3.0, 2.0, 1.0, -1.0, -1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__linear_trend_slope").to_list() == pytest.approx(
        [0.0, 1.0, -1.0, 0.0], abs=1e-6,
    )


def test_has_duplicate_max():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 2.0, 2.0, 1.0, -1.0, -1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__has_duplicate_max").to_list() == [0.0, 0.0, 1.0, 1.0]


def test_has_duplicate_min():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 2.0, 2.0, 1.0, -1.0, -1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__has_duplicate_min").to_list() == [0.0, 0.0, 0.0, 1.0]


def test_has_duplicate1():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 2.0, 2.0, 1.0, -1.0, -1.0, -1.0, -1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__has_duplicate").to_list() == [0.0, 0.0, 1.0, 1.0]


def test_has_duplicate2():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "c", "c", "c", "d", "d", "d"],
            "val": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 1.0, 1.0, 1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__has_duplicate").to_list() == [0.0, 0.0, 0.0, 1.0]


def test_ratio_value_number_to_time_series_length():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 1.0, 1.0, 1.0, 1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column(
        "val__ratio_value_number_to_time_series_length",
    ).to_list() == [1.0, 1.0, 1.0, 0.25]


def test_variation_coefficient():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "c", "c", "c", "d", "d", "d"],
            "val": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 1.0, 2.0, 3.0, -1.0, 0.0, 1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert math.isnan(fdf.get_column("val__variation_coefficient").to_list()[-1])


def test_sum_of_reoccurring_values():
    df = pl.DataFrame(
        {
            "id": [
                "a",
                "a",
                "a",
                "b",
                "b",
                "b",
                "c",
                "c",
                "c",
                "d",
                "d",
                "d",
                "d",
                "d",
                "d",
            ],
            "val": [
                1.0,
                2.0,
                3.0,
                1.0,
                2.0,
                3.0,
                1.0,
                2.0,
                3.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                3.0,
            ],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__sum_of_reoccurring_values").to_list() == [
        0.0,
        0.0,
        0.0,
        3.0,
    ]


def test_sum_of_reoccurring_data_points():
    df = pl.DataFrame(
        {
            "id": [
                "a",
                "a",
                "a",
                "b",
                "b",
                "b",
                "c",
                "c",
                "c",
                "d",
                "d",
                "d",
                "d",
                "d",
                "d",
            ],
            "val": [
                1.0,
                2.0,
                3.0,
                1.0,
                2.0,
                3.0,
                1.0,
                2.0,
                3.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                3.0,
            ],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__sum_of_reoccurring_data_points").to_list() == [
        0.0,
        0.0,
        0.0,
        7.0,
    ]


def test_standard_deviation():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "c"],
            "val": [1.0, 1.0, 1.0, 1.0, 2.0, 3.0, 1.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__standard_deviation").to_list()[0] == 0
    assert math.isnan(fdf.get_column("val__standard_deviation").to_list()[-1])


def test_variance():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "c"],
            "val": [1.0, 1.0, 1.0, 1.0, 2.0, 3.0, 1.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__variance").to_list()[0] == 0
    assert math.isnan(fdf.get_column("val__variance").to_list()[-1])

def test_variance_larger_than_standard_deviation():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "a", "a", "b", "b", "b", "b", "b"],
            "val": [-1.0, -1.0, 1.0, 1.0, 1.0, -1.0, -1.0, 0.1, 0.1, 0.1],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("val__variance_larger_than_standard_deviation").to_list()[0] == 1.0
    assert fdf.get_column("val__variance_larger_than_standard_deviation").to_list()[1] == 0.0


def test_large_standard_deviation():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "a"],
            "val": [-1.0, -1.0, 1.0, 1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("val__large_standard_deviation__r_0.25").to_list()[0] == 1.0
    assert fdf.get_column("val__large_standard_deviation__r_0.30").to_list()[0] == 1.0
    assert fdf.get_column("val__large_standard_deviation__r_0.50").to_list()[0] == 1.0
    assert fdf.get_column("val__large_standard_deviation__r_0.70").to_list()[0] == 0.0

def test_symmetry_looking():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "a", "b", "b", "b", "b", "b", "c", "c", "c", "c", "c", "c", "d", "d"],
            "val": [-1.0, -1.0, 1.0, 1.0, -1.0, -1.0, -1.0, -1.0, 1.0, -2.0, -2.0, -2.0, -1.0, -1.0, -1.0, -0.9, -0.900001],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("val__symmetry_looking__r_0.75").to_list()[0] == 1.0
    assert fdf.get_column("val__symmetry_looking__r_0.05").to_list() == [1.0, 0.0, 1.0, 1.0]

def test_percentage_of_reoccurring_values_to_all_values():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "b", "b", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 1.0, 2.0, 3.0, 3.0, 4.0, 1.0, 1.0, 1.0, -1.0, -1.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column(
        "val__percentage_of_reoccurring_values_to_all_values",
    ).to_list() == [0.5, 0.25, 0.0, 1.0]


def test_percentage_of_reoccurring_values_to_all_datapoints():
    df = pl.DataFrame(
        {
            "id": [
                "a",
                "a",
                "a",
                "a",
                "b",
                "b",
                "b",
                "b",
                "b",
                "c",
                "d",
                "d",
                "d",
                "d",
            ],
            "val": [
                1.0,
                1.0,
                2.0,
                3.0,
                1.0,
                2.0,
                3.0,
                3.0,
                4.0,
                1.0,
                1.0,
                1.0,
                -1.0,
                -1.0,
            ],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column(
        "val__percentage_of_reoccurring_values_to_all_datapoints",
    ).to_list() == pytest.approx([0.25, 0.2, 0.0, 0.5])


def test_agg_linear_trend_intercept():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "b", "b", "b", "b", "b", "b"],
            "val": [1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert math.isnan(
        fdf.get_column(
            "val__agg_linear_trend_intercept__chunk_size_5__agg_mean",
        ).to_list()[0],
    )
    assert fdf.get_column(
        "val__agg_linear_trend_intercept__chunk_size_5__agg_mean",
    ).to_list()[1] == pytest.approx(2.0, abs=1e-6)


def test_agg_linear_trend_slope():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "b", "b", "b", "b", "b", "b"],
            "val": [1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert math.isnan(
        fdf.get_column("val__agg_linear_trend_slope__chunk_size_5__agg_mean").to_list()[
            0
        ],
    )
    assert fdf.get_column(
        "val__agg_linear_trend_slope__chunk_size_5__agg_mean",
    ).to_list()[1] == pytest.approx(1.0, abs=1e-6)


def test_mean_n_absolute_max():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "b", "b", "b", "b", "b"],
            "val": [1.0, 1.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert math.isnan(fdf.get_column("val__mean_n_absolute_max__n_7").to_list()[0])
    assert fdf.get_column("val__mean_n_absolute_max__n_7").to_list()[1] == 4.0


def test_mean_change():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "b", "b", "b", "b", "b"],
            "val": [1.0, 1.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__mean_change").to_list() == [0.0, 1.0]


def test_number_crossing_m1():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "b", "b", "b", "b", "b"],
            "val": [1.0, 1.0, -1.0, 2.0, 3.0, -4.0, -5.0, -6.0, 7.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__number_crossing_m__m_0.0").to_list() == [0.0, 3.0]


def test_number_crossing_m2():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "b", "b", "b", "b", "b"],
            "val": [0.0, 1.0, 0.0, 1.0, -3.0, -4.0, -5.0, 6.0, 7.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__number_crossing_m__m_1.0").to_list() == [0.0, 1.0]


def test_number_crossing_m3():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "b", "b", "b", "b", "b", "b", "b"],
            "val": [-1.0, -1.0, 1.0, 2.0, -3.0, -4.0, -5.0, 6.0, 7.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__number_crossing_m__m_-1.0").to_list() == [0.0, 2.0]


def test_number_crossing_m4():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "b", "b", "b", "b"],
            "val": [1.0, 0.0, -1.0, 1.0, 0.0, 1.0, 0.0, -1.0, 0.0, -1.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__number_crossing_m__m_0.0").to_list() == [1.0, 1.0]


def test_range_count():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "b", "b", "b", "b", "b", "b", "b"],
            "val": [2.0, 0.0, -2.0, 1.0, 0.0, 1.0, 0.0, -2.0, 0.0, -2.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__range_count__min_-1.0__max_1.0").to_list() == [1.0, 5.0]


def test_index_mass_quantile():
    df = pl.DataFrame(
        {
            "id": ["a", "a", "a", "a", "b", "b", "b", "b", "b"],
            "val": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
        },
    ).lazy()

    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")

    assert fdf.get_column("val__index_mass_quantile__q_0.5").to_list() == pytest.approx(
        [0.5, 0.2], abs=1e-6,
    )


def test_c3():
    df = pl.DataFrame(
        {
            "id": ["a"] * 10,
            "val": [1.0] * 10,
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("val__c3__lag_1").to_list() == [1.0]
    assert fdf.get_column("val__c3__lag_2").to_list() == [1.0]
    assert fdf.get_column("val__c3__lag_3").to_list() == [1.0]


def test_c3_2():
    df = pl.DataFrame(
        {
            "id": ["a"] * 4,
            "val": [1, 2, -3, 4],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("val__c3__lag_1").to_list() == [-15.0]
    assert fdf.get_column("val__c3__lag_2").to_list() == [0.0]
    assert fdf.get_column("val__c3__lag_3").to_list() == [0.0]


def test_time_reversal_asymmetry_statistic():
    df = pl.DataFrame(
        {
            "id": ["a"] * 10,
            "val": [1.0] * 10,
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column(
        "val__time_reversal_asymmetry_statistic__lag_1",
    ).to_list() == [0.0]
    assert fdf.get_column(
        "val__time_reversal_asymmetry_statistic__lag_2",
    ).to_list() == [0.0]
    assert fdf.get_column(
        "val__time_reversal_asymmetry_statistic__lag_3",
    ).to_list() == [0.0]


def test_time_reversal_asymmetry_statistic_2():
    df = pl.DataFrame(
        {
            "id": ["a"] * 4,
            "val": [1, 2, -3, 4],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column(
        "val__time_reversal_asymmetry_statistic__lag_1",
    ).to_list() == [-10.0]
    assert fdf.get_column(
        "val__time_reversal_asymmetry_statistic__lag_2",
    ).to_list() == [0.0]
    assert fdf.get_column(
        "val__time_reversal_asymmetry_statistic__lag_3",
    ).to_list() == [0.0]


def test_number_peaks():
    df = pl.DataFrame(
        {
            "id": ["a"] * 14,
            "val": [0, 1, 2, 1, 0, 1, 2, 3, 4, 5, 4, 3, 2, 1],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
    )
    fdf = extract_features(df, opts)
    fdf = fdf.sort("id")
    assert fdf.get_column("val__number_peaks__n_1").to_list() == [2.0]
    assert fdf.get_column("val__number_peaks__n_3").to_list() == [1.0]
    assert fdf.get_column("val__number_peaks__n_5").to_list() == [0.0]



import time

import pandas as pd
import polars as pl
import tsfresh
import tsfx
from tsfresh.feature_extraction.settings import MinimalFCParameters
from tsfx import ExtractionSettings, FeatureSetting


def tsfresh_minimal() -> None:
    """Benchmark TSFRESH minimal features extraction."""
    print("Benchmarking TSFRESH minimal features extraction...")
    print("Loading data...")
    df = pd.read_csv("./test_data/all_stocks_5yr.csv")
    df = df.dropna(axis=0, how="any")
    print("Data loaded")
    settings = MinimalFCParameters()
    print("Starting TSFRESH minimal features extraction...")
    start = time.time()
    fdf = tsfresh.extract_features(
        df,
        column_id="Name",
        column_sort="date",
        default_fc_parameters=settings,
    )
    end = time.time()
    print(f"TSFRESH Minimal extraction took {end - start} s")
    print(fdf.head())


def tsfx_minimal() -> None:
    """Benchmark TSFX minimal features extraction."""
    print("Benchmarking TSFX minimal features extraction...")
    print("Loading data...")
    lf = pl.scan_csv("test_data/all_stocks_5yr.csv")
    lf = lf.drop_nulls()
    print("Data loaded")
    opts = ExtractionSettings(
        grouping_col="Name",
        feature_setting=FeatureSetting.Minimal,
        value_cols=["open", "high", "low", "close", "volume"],
    )
    print("Starting TSFX minimal features extraction...")
    start = time.time()
    fdf = tsfx.extract_features(lf, opts)
    end = time.time()
    print(f"TSFX Minimal extraction took {end - start} s")
    print(fdf.sort(pl.col("Name")).head())


if __name__ == "__main__":
    tsfx_minimal()
    tsfresh_minimal()



import json
import time

import pandas as pd
import polars as pl
import tsfresh
import tsfx
from tsfx import ExtractionSettings, FeatureSetting


def tsfresh_efficient() -> float:
    """Benchmark TSFRESH efficient features extraction."""
    print("Benchmarking TSFRESH efficient features extraction...")
    print("Loading data...")
    df = pd.read_csv("./test_data/all_stocks_5yr.csv")
    df = df.dropna(axis=0, how="any")
    print("Data loaded")
    with open("./python_benchmarks/efficient.json") as f:
        settings = json.load(f)
    print("Starting TSFRESH efficient features extraction...")
    start = time.time()
    fdf = tsfresh.extract_features(
        df[
            [
                "date",
                "open",
                "high",
                "low",
                "close",
                "volume",
                "Name",
            ]
        ],
        column_id="Name",
        column_sort="date",
        default_fc_parameters=settings,
    )
    end = time.time()
    print(f"TSFRESH Efficient extraction took {end - start} s")
    print(fdf.head())
    return end - start


def tsfx_efficient() -> float:
    """Benchmark TSFX efficient features extraction."""
    print("Benchmarking TSFX efficient features extraction...")
    print("Loading data...")
    lf = pl.scan_csv("test_data/all_stocks_5yr.csv")
    lf = lf.drop_nulls()
    print("Data loaded")
    opts = ExtractionSettings(
        grouping_col="Name",
        feature_setting=FeatureSetting.Efficient,
        value_cols=[
            "open",
            "high",
            "low",
            "close",
            "volume",
        ],
    )
    print("Starting TSFX efficient features extraction...")
    start = time.time()
    fdf = tsfx.extract_features(lf, opts)
    end = time.time()
    print(f"TSFX Efficient extraction took {end - start} s")
    print(fdf.sort(pl.col("Name")).head())
    return end - start


if __name__ == "__main__":
    time1 = tsfx_efficient()
    time2 = tsfresh_efficient()
    print(f"Speed-up ~ {time2/time1:.1f}x")



from __future__ import annotations

import polars as pl

from tsfx import tsfx


def extract_features(
    lf: pl.LazyFrame,
    settings: tsfx.ExtractionSettings,
    streaming: bool = False,
) -> pl.DataFrame:
    fdf = tsfx.extract_features(
        lf=lf,
        settings=settings,
        streaming=streaming,
    )

    return fdf.unnest(
        [
            col_name
            for col_name, dtype in zip(
                fdf.columns,
                fdf.dtypes,
                strict=True,
            )
            if dtype == pl.Struct
        ],
    )



import polars as pl
from tsfx import (
    ExtractionSettings,
    FeatureSetting,
    extract_features,
)


def test_empty_config():
    df = pl.DataFrame(
        {
            "id": ["a", "b", "b", "c", "c", "c", "d", "d", "d", "d"],
            "val": [1.0, 1.0, 2.0, 1.0, 2.0, 3.0, 1.0, 1.0, 1.0, 1.0],
        },
    ).lazy()
    opts = ExtractionSettings(
        grouping_col="id",
        feature_setting=FeatureSetting.Efficient,
        value_cols=["val"],
        config_path="./python_tests/data/.tsfx-config-empty.toml",
    )
    gdf = extract_features(df, opts)
    print(gdf.head())
    assert gdf.shape == (4, 1)



import polars as pl
from tsfx import (
    DynamicGroupBySettings,
    ExtractionSettings,
    FeatureSetting,
    extract_features,
)

lf = pl.scan_csv("test_data/all_stocks_5yr.csv")
lf = lf.drop_nulls()

dyn_opts = DynamicGroupBySettings(
    time_col="date",
    every="1y",
    period="1y",
    offset="0",
    datetime_format="%Y-%m-%d",
)

opts = ExtractionSettings(
    grouping_col="Name",
    value_cols=["open", "high", "low", "close", "volume"],
    feature_setting=FeatureSetting.Efficient,
    dynamic_settings=dyn_opts,
)
gdf = extract_features(lf, opts)
print(gdf.sort(by=pl.col("Name")))

opts = ExtractionSettings(
    grouping_col="Name",
    feature_setting=FeatureSetting.Efficient,
    value_cols=["open", "high", "low", "close", "volume"],
)
gdf = extract_features(lf, opts)
print(gdf.sort(by=pl.col("Name")))



import polars as pl
from tsfx import (
    DynamicGroupBySettings,
    ExtractionSettings,
    FeatureSetting,
    extract_features,
)

df = pl.DataFrame(
    {
        "id": ["a", "a", "a", "b", "b", "b", "c", "c", "c"],
        "val": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
        "value": [4.0, 5.0, 6.0, 6.0, 5.0, 4.0, 4.0, 5.0, 6.0],
    },
).lazy()
settings = ExtractionSettings(
    grouping_col="id",
    feature_setting=FeatureSetting.Efficient,
    value_cols=["val", "value"],
)
gdf = extract_features(df, settings)
gdf = gdf.sort(by="id")
print(gdf)

tdf = pl.DataFrame(
    {
        "id": ["a", "a", "a", "b", "b", "b", "c", "c", "c"],
        "time": [
            "2001-01-01",
            "2002-01-01",
            "2003-01-01",
            "2001-01-01",
            "2002-01-01",
            "2003-01-01",
            "2001-01-01",
            "2002-01-01",
            "2003-01-01",
        ],
        "val": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
        "value": [4.0, 5.0, 6.0, 6.0, 5.0, 4.0, 4.0, 5.0, 6.0],
    },
).lazy()

dyn_settings = DynamicGroupBySettings(
    time_col="time",
    every="3y",
    period="3y",
    offset="0",
    datetime_format="%Y-%m-%d",
)
settings = ExtractionSettings(
    grouping_col="id",
    value_cols=["val", "value"],
    feature_setting=FeatureSetting.Efficient,
    dynamic_settings=dyn_settings,
)
gdf = extract_features(tdf, settings)
gdf = gdf.sort(by="id")
print(gdf)



from __future__ import annotations

from collections.abc import Sequence
from enum import Enum, auto

class FeatureSetting(Enum):
    Minimal = auto()
    Efficient = auto()
    Comprehensive = auto()

class ExtractionSettings:
    def __init__(
        self,
        grouping_col: str,
        value_cols: Sequence[str],
        feature_setting: FeatureSetting,
        config_path: str | None = None,
        dynamic_settings: DynamicGroupBySettings | None = None,
    ) -> None:
        self.grouping_col = grouping_col
        self.value_cols = value_cols
        self.feature_setting = feature_setting
        self.dynamic_settings = dynamic_settings

class DynamicGroupBySettings:
    def __init__(
        self,
        time_col: str,
        every: str,
        period: str,
        offset: str,
        datetime_format: str,
    ) -> None:
        self.time_col = time_col
        self.every = every
        self.period = period
        self.offset = offset
        self.datetime_format = datetime_format
